/* 
 * tkGeomManager.c --
 *
 *	This file contains code generic Tk code for geometry
 *	menegers.
 *
 * Author: Pekka Nikander <pnr@innopoli.ajk.tele.fi>
 *
 * Copyright 1992 Telecom Finland
 * Permission to use, copy, modify, and distribute this
 * software and its documentation for any purpose and without
 * fee is hereby granted, provided that the above copyright
 * notice appear in all copies.  Telecom Finland
 * makes no representations about the suitability of this
 * software for any purpose.  It is provided "as is" without
 * express or implied warranty.
 *
 * Created: Wed Mar 11 17:20:55 1992
 * Last modified: Wed Mar 11 20:53:14 1992 pnr
 */

#ifndef lint
static char rcsid[] = "...";
#endif

#include <assert.h>

#include "tkconfig.h"
#include "tkint.h"
#include "tkgeommanager.h"

static void GeomManagerStructureProc _ANSI_ARGS_ ((ClientData clientData,
			      Tk_Window tkwin));

/*
 *--------------------------------------------------------------
 *
 * TkCreateGeomManager --
 *
 *	This internal procedure is used to create a geometry manager
 *	structure for a given window.
 *
 * Results:
 *	The return value is a pointer to a geometry manager structure
 *	corresponding to tkwin.
 *
 * Side effects:
 *	A new geometry manager structure is created. 
 *	A callback is set up to clean things up when the
 *	window is deleted.
 *
 *--------------------------------------------------------------
 */

TkGeomManager *
TkCreateGeomManager(interp, tkwin, name, arrangeProc, destroyProc, clientData)
    Tcl_Interp *interp;
    Tk_Window tkwin;		/* Token for window for which
				 * Tk_Geometry structure is desired. */
    char *name;			/* Name of the geometry manager */
    void (*arrangeProc)();	/* Procedure to rearrange the manager. */
    void (*destroyProc)();	/* Procedure to cleanup.  Called from
				 * DestroyGeomManager. */
    ClientData *clientData;	/* Argument to arrangeProc and destroyProc. */
{
    register TkWindow *winPtr = (TkWindow *) tkwin;
    register TkGeomManager *geomPtr;

    geomPtr = (TkGeomManager *) ckalloc(sizeof(TkGeomManager));

    geomPtr->display = winPtr->display;
    geomPtr->dispPtr = winPtr->dispPtr;
    geomPtr->screenNum = winPtr->screenNum;
    geomPtr->winFlags = 0;

    NameWindow(interp, (TkWindow *)geomPtr, winPtr, name);
    Tk_SetClass((TkWindow *)geomPtr, "GeomManager");
    geomPtr->optionLevel = -1;

    geomPtr->childPtr = NULL;
    geomPtr->parentPtr = NULL;
    geomPtr->nextPtr = NULL;
    geomPtr->flags = 0;
    geomPtr->arrangeProc = arrangeProc;
    geomPtr->clientData = clientData;

    Tk_CreateEventHandler(tkwin, StructureNotifyMask,
			 GeomManagerStructureProc, (ClientData) geomPtr);
    return geomPtr;
}

/*
 *--------------------------------------------------------------
 *
 * TkLinkGeomManager --
 *
 *	This procedure links a child geometry manager
 *	at a proper place into the managed set of the
 *	parent manager.
 *
 * Results:
 *	A standard Tcl return value.
 *
 * Side effects:
 *	The location of the specified windows may change.
 *
 *--------------------------------------------------------------
 */

void
TkLinkGeomManager(parentPtr, childPtr, prevPtr)
    TkGeomManager *parentPtr;	/* Parent in which to show windows. */
    TkGeomManager *childPtr;	/* Child to link. */
    TkGeomManager *prevPtr;	/* Previous child in link order, or NULL. */
{
    if (childPtr != prevPtr) {
	/*
	 * Unshow this window if it's currently shown.
	 */
	if (childPtr->parentPtr != NULL) {
	    TkUnlinkGeomManager(childPtr);
	}
	
	/*
	 * Add the window in the correct place in its parent's
	 * show order, then make sure that the window is
	 * managed by us.
	 */
	
	childPtr->parentPtr = parentPtr;
	if (prevPtr == NULL) {
	    childPtr->nextPtr = parentPtr->childPtr;
	    parentPtr->childPtr = childPtr;
	} else {
	    childPtr->nextPtr = prevPtr->nextPtr;
	    prevPtr->nextPtr = childPtr;
	}
    }

    /*
     * Arrange for the parent to be rearranged at the first
     * idle moment.
     */

    if (!(parentPtr->flags & REQUESTED_REARRANGE)) {
	parentPtr->flags |= REQUESTED_REARRANGE;
	Tk_DoWhenIdle(parentPtr->arrangeProc, parentPtr->clientData);
    }
}

/*
 *--------------------------------------------------------------
 *
 * TkGeomManagerReqProc --
 *
 *	This procedure is invoked by Tk_GeometryRequest for
 *	windows managed by a geometry manager.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Arranges for tkwin, and all its managed siblings, to
 *	be rearranged at the next idle point.
 *
 *--------------------------------------------------------------
 */

	/* ARGSUSED */
void
TkGeomManagerReqProc(clientData, tkwin)
    ClientData clientData;	/* Packer's information about
				 * window that got new preferred
				 * geometry.  */
    Tk_Window tkwin;		/* Other Tk-related information
				 * about the window. */
{
    register TkGeomManager *geomPtr = (TkGeomManager *) clientData;

    assert(geomPtr->parentPtr != NULL);
    geomPtr = geomPtr->parentPtr;
    if (!(geomPtr->flags & REQUESTED_REARRANGE)) {
	geomPtr->flags |= REQUESTED_REARRANGE;
	Tk_DoWhenIdle(geomPtr->arrangeProc, geomPtr->clientData);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * TkUnlinkGeomManager --
 *
 *	Remove a geometry manager from its parent's list of children.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The parent will be scheduled for rearranging.
 *
 *----------------------------------------------------------------------
 */

void
TkUnlinkGeomManager(childPtr)
    register TkGeomManager *childPtr;		/* Window to unlink. */
{
    register TkGeomManager *parentPtr, *childPtr2;

    parentPtr = childPtr->parentPtr;
    if (parentPtr == NULL) {
	return;
    }
    if (parentPtr->childPtr == childPtr) {
	parentPtr->childPtr = childPtr->nextPtr;
    } else {
	for (childPtr2 = parentPtr->childPtr;
	     ;
	     childPtr2 = childPtr2->nextPtr) {
	    if (childPtr2 == NULL) {
		panic("Unlink couldn't find previous window");
	    }
	    if (childPtr2->nextPtr == childPtr) {
		childPtr2->nextPtr = childPtr->nextPtr;
		break;
	    }
	}
    }
    if (!(parentPtr->flags & REQUESTED_REARRANGE)) {
	parentPtr->flags |= REQUESTED_REARRANGE;
	Tk_DoWhenIdle(parentPtr->arrangeProc, parentPtr->clientData);
    }
    childPtr->parentPtr = NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * DestroyGeomManager --
 *
 *	This procedure is invoked by Tk_EventuallyFree or Tk_Release
 *	to clean up the internal structure of a geomery manager at a safe time
 *	(when no-one is using it anymore).
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Everything associated with the geometry manager is freed up.
 *
 *----------------------------------------------------------------------
 */

static void
DestroyGeomManager(clientData)
    ClientData clientData;		/* Manager info about window that
					 * is now dead. */
{
    register TkGeomManager *geomPtr = (TkGeomManager *) clientData;
    TkUnnameWindow((Tk_Window) geomPtr);
    if (geomPtr->destroyProc != NULL) {
	(*geomPtr->destroyProc)(geomPtr->clientData);
    }
    ckfree((char *) geomPtr);
}


/*
 *----------------------------------------------------------------------
 *
 * GeomManagerStructureProc --
 *
 *	This procedure is invoked by the Tk event dispatcher in response
 *	to StructureNotify events.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	If a window was just deleted, clean up all its geometry
 *	manager related information.  If it was just resized, rearrange
 *	its children, if any.
 *
 *----------------------------------------------------------------------
 */

static void
GeomManagerStructureProc(clientData, eventPtr)
    ClientData clientData;		/* Our information about window
					 * referred to by eventPtr. */
    XEvent *eventPtr;			/* Describes what just happened. */
{
    register TkGeomManager *geomPtr = (TkGeomManager *) clientData;
    if (eventPtr->type == ConfigureNotify) {
	if ((geomPtr->childPtr != NULL)
		&& !(geomPtr->flags & REQUESTED_REARRANGE)) {
	    geomPtr->flags |= REQUESTED_REARRANGE;
	    Tk_DoWhenIdle(geomPtr->arrangeProc, geomPtr->clientData);
	}
    } else if (eventPtr->type == DestroyNotify) {
	register TkGeomManager *geomPtr2;

	if (geomPtr->parentPtr != NULL) {
	    TkUnlinkGeomManager(geomPtr);
	}
	for (geomPtr2 = geomPtr->childPtr; geomPtr2 != NULL;
		geomPtr2 = geomPtr2->nextPtr) {
	    geomPtr2->parentPtr = NULL;
	    geomPtr2->nextPtr = NULL;
	}
	if (geomPtr->flags & REQUESTED_REARRANGE) {
	    Tk_CancelIdleCall(geomPtr->arrangeProc, geomPtr->clientData);
	}
	geomPtr->tkwin = NULL;
	Tk_EventuallyFree((ClientData) geomPtr, DestroyGeomManager);
    }
}
